/*
    Copyright (C) 2007 James Bourne <jbourne@hardrock.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <assert.h>
#include <limits.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>

char progname[] = "check_file";
char version[] = "$Id: check_file.c,v 1.5 2007/03/03 19:05:20 jbourne Exp jbourne $";

struct val {
	time_t warn;
	time_t crit;
	int exist;
	int invert;
	char path[PATH_MAX+1];
} vals;
	
void display_help(void)
{
	printf("\n\nHELP!\n%s version %s\n", progname, version);
	printf("%s is a nagios plugin that you can use to monitor a file has changed\n", progname);
	printf("within N seconds, where N is an option to critical or warning\n\n");
	printf("Options:\n -w, --warning=seconds\n\tseconds since file last changed to warn\n");
	printf(" -c, --critical=seconds\n\tseconds since file last changed to issue critical\n");
	printf(" -p, --path=path\n\tpath to file to monitor\n");
	printf(" -e, --exist\n\tcheck for existance of the file only, can't be used with age.\n");
	printf(" -i, --invert\n\tdo the opposite check, file older than age, or file does not exist\n");
	printf(" -v, --version\n\tversion of %s\n", progname);
	printf(" -h, --help\n\t\tprint this help\n\n");
}

int main(int argc, char *argv[])
{
	/* get the options */
	int option = 0;
	int c;
	struct stat fstat;
	time_t now;
	
	vals.crit = -1;
	vals.warn = -1;
	vals.invert = 0;
	vals.exist = 0;
	memset(vals.path, '\0', PATH_MAX+1);

	static struct option longopts[] = {
		{"warning", required_argument, 0, 'w'},
		{"critical", required_argument, 0, 'c'},
		{"path", required_argument, 0, 'p'},
		{"exist", no_argument, 0, 'e'},
		{"invert", no_argument, 0, 'i'},
		{"version", no_argument, 0, 'V'},
		{"verbose", no_argument, 0, 'v'},
		{"help", no_argument, 0, 'h'},
		{0, 0, 0, 0}
	};

	if (argc < 2) {
		printf("SERVICE STATUS: UNKNOWN Please check arguments\n");
		display_help();
		exit (3);
	}
	
	while(1) {
		c = getopt_long (argc, argv, "+?Vqhveic:w:p:", longopts, &option);
		if(c == -1 || c == EOF)
			break;
		
		switch(c) {
			case 'w':
					if(optarg) 
						vals.warn = atol(optarg);
				break;
			case 'c':
					if(optarg)
						vals.crit = atol(optarg);
				break;
			case 'e':
						vals.exist = 1;
				break;
			case 'i':
						vals.invert = 1;
				break;
			case 'p':
					if(optarg)
						strncpy(vals.path, optarg, PATH_MAX);
				break;
			case 'V':
					printf("VERSION: %s: %s\n", progname, version);
					exit(0);
				break;
			case 'v':
				/* ignored */
				break;
			case 'h':
			case '?':
			default:
				printf("SERVICE STATUS: UNKNOWN Please check arguments\n");
				display_help();
				exit(3);
		}
	}	

	if(vals.exist && ((vals.crit != -1)  || (vals.warn != -1))) {
		printf("SERVICE STATUS: UNKNOWN You can not use exist and critical or warning ages\n");
		display_help();
		exit(3);
	}

	if(strlen(vals.path) == 0) {
		printf("SERVICE STATUS: You must provide a valid file path\n");
		display_help();
		exit(-1);
	}
	
	if(vals.exist) {
		/* check existance only */	
		if(stat(vals.path, &fstat)) {
			if(errno == ENOENT) {
				if(vals.invert) {
					printf("SERVICE STATUS: OK file or directory %s does not exist\n", vals.path);
					exit(0);
				}

				printf("SERVICE STATUS: CRITICAL file or directory %s does not exist\n", vals.path);
				exit(2);
			}
			printf("SERVICE STATUS: CRITICAL An error occured: %s\n", strerror(errno));
			display_help();
			exit(2);
		}
		
		/* file exists */
		if(vals.invert) {
			printf("SERVICE STATUS: CRITICAL file or directory %s exists\n", vals.path);
			exit(2);
		}
		printf("SERVICE STATUS: OK file or directory %s exists\n", vals.path);
		exit(0);
	}

	if((vals.crit < 0) || (vals.warn < 0)) {
		printf("SERVICE STATUS: UNKNOWN You must provide warning or critical ages greater than 0 seconds\n");
		display_help();
		exit(3);
	}

	if(vals.invert) {
		if(vals.crit > vals.warn) {
			printf("SERVICE STATUS: UNKNOWN Your critical value must be larger than your warning value\n");
			display_help();
			exit(3);
		}
	} else {
		if(vals.crit < vals.warn) {
			printf("SERVICE STATUS: UNKNOWN Your critical value must be larger than your warning value\n");
			display_help();
			exit(3);	
		}
	}
	
	/* stat the file */
	if(stat(vals.path, &fstat)) {
		/* print errno */
		printf("SERVICE STATUS: CRITICAL An error occured: %s\n", strerror(errno));
		display_help();
		exit(2);
	}

	time(&now);
	now = mktime(localtime(&now));	
	/* compare to given age */

	if(vals.invert) {
		if(now - (time_t)vals.crit <= fstat.st_mtime) {
			/* print a critical */
			printf("SERVICE STATUS: CRITICAL file %s changed within %ld seconds\n", vals.path, now-fstat.st_mtime);
			exit(2);
		}

		if(now - (time_t)vals.warn <= fstat.st_mtime) {
			/* print a warning */
			printf("SERVICE STATUS: WARNING file %s changed within %ld seconds\n", vals.path, now-fstat.st_mtime);
			exit(1);
		}

		printf("SERVICE STATUS: OK file %s has not changed for over %ld seconds\n", vals.path, now-fstat.st_mtime);
		exit(0);
	}

	if(now - (time_t)vals.crit > fstat.st_mtime) {
		/* print a critical */
		printf("SERVICE STATUS: CRITICAL file %s not changed in %ld seconds\n", vals.path, now-fstat.st_mtime);
		exit(2);
	}

	if(now - (time_t)vals.warn > fstat.st_mtime) {
		/* print a warning */
		printf("SERVICE STATUS: WARNING file %s not changed in %ld seconds\n", vals.path, now-fstat.st_mtime);
		exit(1);
	}

	printf("SERVICE STATUS: OK file %s changed %ld seconds ago\n", vals.path, now-fstat.st_mtime);
	exit(0);
}
