#!/bin/sh
#
# Nagios plugin to check Oracle tablespace usage.
#
# $Id: check_oracle_tablespace.sh,v 1.7 2007/02/16 09:41:16 kivimaki Exp $
#
# Copyright (C) 2006  Hannu Kivimki / CSC - Scientific Computing Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

# Oracle environment settings (could be parametrized)
ORACLE_ORATAB="/path/to/oratab"
ORACLE_USER="username"
ORACLE_PASS="password"

# External commands
CMD_AWK="/usr/bin/awk"
CMD_EGREP="/usr/bin/egrep"

# Temporary work file (will be removed automatically)
TEMP_FILE="/tmp/check_oracle_tablespace_$$.tmp"

# Nagios plugin return values
RET_OK=0
RET_WARNING=1
RET_CRITICAL=2
RET_UNKNOWN=3

# Default values
THRESHOLD_ERROR=0
WARNING_THRESHOLD=-1
WARNING_EXCEEDED=0
WARNING_STATUS_TEXT=""
CRITICAL_THRESHOLD=-1
CRITICAL_EXCEEDED=0
CRITICAL_STATUS_TEXT=""
CHECK_AUTOEXTENSION=0
VERBOSE=0

# ------------------------------ FUNCTIONS -------------------------------------

printInfo() {
    echo "Nagios plugin to check Oracle tablespace usage."
    echo "Copyright (C) 2006  Hannu Kivimki / CSC - Scientific Computing Ltd."
}

printHelp() {
    echo
    echo "Usage: check_oracle_tablespace.sh -s SID [-d <regexp>] -w <1-100> -c <1-100> [-a]"
    echo
    echo "  -s  Oracle system identifier (SID)"
    echo "  -d  which tablespaces/databases to check (defaults to all)"
    echo "  -w  warning threshold (usage% as integer)"
    echo "  -c  critical threshold (usage% as integer)"
    echo "  -a  check autoextension - autoextensible tablespaces will return"
    echo "      only a warning even if critical threshold is exceeded. All"
    echo "      autoextensible tablespaces are also marked with AUTOEXT in"
    echo "      status text."
    echo "  -h  this help screen"
    echo "  -l  license info"
    echo "  -v  verbose output (for debugging)"
    echo "  -V  version info"
    echo
    echo "Example: check_oracle_tablespace.sh -d 'FOO.*' -w 90 -c 95"
    echo
    echo "This will return CRITICAL if tablespace usage of any database"
    echo "matching regular expression 'FOO.*' (case insensitive, will be"
    echo "used as a parameter for 'egrep -i') is 95% or more, WARNING if"
    echo "usage is 90% or more and otherwise OK. Warning threshold is ignored"
    echo "if it is higher or equal to critical threshold."
    echo
}

printLicense() {
    echo
    echo "This program is free software; you can redistribute it and/or"
    echo "modify it under the terms of the GNU General Public License"
    echo "as published by the Free Software Foundation; either version 2"
    echo "of the License, or (at your option) any later version."
    echo
    echo "This program is distributed in the hope that it will be useful,"
    echo "but WITHOUT ANY WARRANTY; without even the implied warranty of"
    echo "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the"
    echo "GNU General Public License for more details."
    echo
    echo "You should have received a copy of the GNU General Public License"
    echo "along with this program; if not, write to the Free Software"
    echo "Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."
    echo
}

printVersion() {
    echo
    echo "\$Id: check_oracle_tablespace.sh,v 1.7 2007/02/16 09:41:16 kivimaki Exp $"
    echo
}

# Checks command line options (pass $@ as parameter).
checkOptions() {
    if [ $# -eq 0 ]; then
        printInfo
        printHelp
        exit $RET_UNKNOWN
    fi

    while getopts s:d:w:c:alhvV OPT $@; do
            case $OPT in
                s) # Oracle SID
                   ORACLE_SID="$OPTARG"
                   ;;
                d) # Oracle databases (regular expression for egrep)
                   DB_REGEXP="$OPTARG"
                   ;;
                w) # warning threshold
                   WARNING_THRESHOLD=$OPTARG
                   ;;
                c) # critical threshold
                   CRITICAL_THRESHOLD=$OPTARG
                   ;;
                a) # check tablespace autoextension
                   CHECK_AUTOEXTENSION=1
                   ;;
                l) printInfo
                   printLicense
                   exit $RET_OK
                   ;;
                h) printInfo
                   printHelp
                   exit $RET_OK
                   ;;
                v) VERBOSE=1
                   ;;
                V) printInfo
                   printVersion
                   exit $RET_OK
                   ;;
                ?) printInfo
                   printHelp
                   exit $RET_UNKNOWN
                   ;;
            esac
    done

    TMP=`$CMD_EGREP \^${ORACLE_SID}\: $ORACLE_ORATAB`
    if [ -z "$ORACLE_SID" ] || [ "$TMP" = "" ]; then
        echo "Error: Invalid Oracle SID (see $ORACLE_ORATAB)."
        printInfo
        printHelp
        exit $RET_UNKNOWN
    else
        ORACLE_HOME=`echo $TMP | $CMD_AWK 'BEGIN{FS=":"}{print $2}'`
    fi

    if [ -z "$DB_REGEXP" ]; then
        DB_REGEXP=".*"
    fi

    if [ $WARNING_THRESHOLD -lt 0 ] || [ $WARNING_THRESHOLD -gt 100 ]; then
        THRESHOLD_ERROR=1
    elif [ $CRITICAL_THRESHOLD -lt 0 ] || [ $CRITICAL_THRESHOLD -gt 100 ]; then
        THRESHOLD_ERROR=1
    fi
    if [ $THRESHOLD_ERROR -eq 1 ]; then
        echo "Error: Threshold values must be between 0-100."
        printInfo
        printHelp
        exit $RET_UNKNOWN
    fi
}

# ----------------------------- MAIN PROGRAM -----------------------------------

checkOptions $@

if [ $VERBOSE -eq 1 ]; then
    echo "ORACLE_SID = $ORACLE_SID"
    echo "ORACLE=HOME = $ORACLE_HOME"
fi
export ORACLE_SID
export ORACLE_HOME

# Feed SQL as here document to sqlplus to query tablespace information.
# Result is stored into temporary text file with three columns (tablespace name,
# usage percentage as integer, autoextensible) separated with whitespace:
#
#  SOMETABLESPACE1          74 YES
#  SOMETABLESPACE2          90 NO
#  ...
#
if [ $VERBOSE -eq 1 ]; then
    echo "Executing $ORACLE_HOME/bin/sqlplus..."
fi
if [ ! -x "$ORACLE_HOME/bin/sqlplus" ]; then
    echo "Error: $ORACLE_HOME/bin/sqlplus not found or not executable."
    exit $RET_UNKNOWN
fi
$ORACLE_HOME/bin/sqlplus $ORACLE_USER/$ORACLE_PASS <<EOF | $CMD_EGREP -i "$DB_REGEXP" | $CMD_EGREP "YES$|NO$" > $TEMP_FILE
set linesize 50
set pages 500
set head off

column tablespace_name format a20
column usage_pct       format 999
column autoextensible  format a5

break on report

select a.tablespace_name,
       ((1 - b.freeb/a.abytes)*100) usage_pct,
       autoextensible
from (select tablespace_name, sum(bytes) as abytes, autoextensible
      from dba_data_files group by tablespace_name, autoextensible) a,
     (select tablespace_name, sum(bytes) as freeb
      from dba_free_space group by tablespace_name) b
where a.tablespace_name = b.tablespace_name
order by a.tablespace_name
/
EOF
if [ "`cat $TEMP_FILE`" = "" ]; then
    echo "Error: Empty result from sqlplus. Check plugin settings and Oracle status."
    exit $RET_UNKNOWN
fi
if [ $VERBOSE -eq 1 ]; then
    cat $TEMP_FILE
fi

# Loop through tablespace usage percentages and set a flag if thresholds
# are exceeded.
#
if [ $VERBOSE -eq 1 ]; then
    echo "Comparing usage percentages to threshold values..."
fi
column=0
for row in `cat $TEMP_FILE`; do
    column=`expr $column + 1`
    case $column in
        1) # tablespace name
           ts=$row
           ;;
        2) # usage percentage
           usage=$row
           ;;
        3) # autoextensible
           autoext=$row
           if [ $usage -ge $WARNING_THRESHOLD ]; then
              # Warning threshold (and perhaps critical as well) was exceeded.
              if [ $CHECK_AUTOEXTENSION -eq 1 ] && [ $autoext = "YES" ]; then
                  # Option '-a' was given and this is an autoextensible tablespace.
                  # Append tablespace, usage and autoextension to status text
                  # (shown in Nagios service status information).
                  WARNING_EXCEEDED=1
                  WARNING_STATUS_TEXT="${WARNING_STATUS_TEXT}${ts} AUTOEXT ${usage}% "
                  if [ $VERBOSE -eq 1 ]; then
                    echo "${ts} ${usage}% WARNING (threshold(s) exceeded but autoextensible)"
                  fi
              elif [ $usage -ge $CRITICAL_THRESHOLD ]; then
                  # Critical threshold was exceeded and option '-a' was not given.
                  # Append tablespace and usage to status text (shown in Nagios
                  # service status information).
                  CRITICAL_EXCEEDED=1
                  CRITICAL_STATUS_TEXT="${CRITICAL_STATUS_TEXT}${ts} ${usage}% "
                  if [ $VERBOSE -eq 1 ]; then
                      echo "${ts} ${usage}% CRITICAL"
                  fi
              else
                  # Warning threshold (but not critical) was exceeded and option
                  # '-a' was not given. Append tablespace and usage to status text
                  # (shown in Nagios service status information).
                  WARNING_EXCEEDED=1
                  WARNING_STATUS_TEXT="${WARNING_STATUS_TEXT}${ts} ${usage}% "
                  if [ $VERBOSE -eq 1 ]; then
                      echo "${ts} ${usage}% WARNING"
                  fi
              fi
           fi
           column=0
           ;;
    esac
done

# Remove temporary work file.
rm -f $TEMP_FILE

# Print check results and exit.
if [ $CRITICAL_EXCEEDED -eq 1 ]; then
    if [ $WARNING_EXCEEDED -eq 1 ]; then
        echo "TABLESPACE CRITICAL: $CRITICAL_STATUS_TEXT WARNING: $WARNING_STATUS_TEXT"
    else
        echo "TABLESPACE CRITICAL: $CRITICAL_STATUS_TEXT"
    fi
    exit $RET_CRITICAL
elif [ $WARNING_EXCEEDED -eq 1 ]; then
    echo "TABLESPACE WARNING: $WARNING_STATUS_TEXT"
    exit $RET_WARNING
else
    echo "TABLESPACE OK"
    exit $RET_OK
fi
